/*
 * File         : BV_Session_Control.cpp
 * Author       : Tomasz Stachowiak
 * Created      : Thu May 21 20:31:07 1998
 * Revision     : $Revision$
 * Updated      : $Date$
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

#include <map.h>
#include <semaphore.h>

#include <BVN/Message.h>
#include <BVN/Reactor.h>
#include <BVN/SOCK_Library.h>

#include "UserData.h"
#include "BV_Member.h"
#include "BV_Msg_Header.h"
#include "BV_Session_Control.h"
#include "BV.h"

//---GUI callbacks
#include "MainForm.h"
#include "StatForm.h"
#include "MainWindowMainWindow.h"
#include "StatWindowMainWindow.h"
//---GUI callbacks END

int BV_Session_Control::reference_no_ = 0;
BV_Session_Control *BV_Session_Control::session_control_ = 0;

// Singleton initialization
BV_Session_Control *BV_Session_Control::instance()
{
  if(reference_no_ == 0)
    {
      session_control_ = new BV_Session_Control();
    }

  reference_no_++;

  return session_control_;
}

// Singleton destruction
void BV_Session_Control::destroy()
{
  if(reference_no_ > 1)
    {
      reference_no_--;
      return;
    }

  if(reference_no_ == 1)
    {
      delete session_control_;
      session_control_ = 0;
      reference_no_--;
      return;
    }

  if(reference_no_ == 0)
    {
      session_control_ = 0;
      return;
    }
}

// Constructor
BV_Session_Control::BV_Session_Control()
{
  BVN_DEBUG("BV_Session_Control::BV_Session_Control\n");
  reference_no_ = 0;
  session_control_ = 0;

  status_ = _init;
  inviting_member_ = 0;
  new_member_ = 0;
  local_id_ = ID_NULL;
  last_id_ = -1;
  internal_id_ = -1;

  sem_init(&sem_, 0, 1);

  if(restore_options() == -1)
    {
      options_.connection = _modem28;
      options_.auto_accept = True;
      strcpy(options_.local_name, "user");
      strcpy(options_.email, "");
      strcpy(options_.server_name, "kopernik.npac.syr.edu");
      options_.manual_switch = 0;

      options_.volume = 125;
      options_.mic_volume = 125;

      options_.fps_max = 5;

      options_.mic_cutoff = 0;
    }
}

BV_Session_Control::~BV_Session_Control()
{
  save_options();
}

void BV_Session_Control::set_reactor(Reactor *reactor)
{
  reactor_ = reactor;
}

////////////////////////////////////////////////////////////////////
//
// Network callbacks
//
int BV_Session_Control::invited(BV_Member *member)
{
  if(status_ == _init)
    {
      inviting_member_ = member;
      status_ = _invited;

      char str[100];
      char str1[150];
      inviting_member_->get_string(str);
      strcpy(str1, "You are INVITED by ");
      strcat(str1, str);
      //---GUI callbacks
      form_->setStatus(str1);
      form_->activateButtons(True);
      //---GUI callbacks END

      if(options_.auto_accept)
	{
	  accept_invitation();
	  form_->activateButtons(False);
	} 
    }
  else
    {
      BV_Msg_Header header;
      header.set_tag(reqRemoteSessionReject);
      BV_Message msg(header);
      member->send_msg(&msg);
      delete member;
    }

  return 0;
}

int BV_Session_Control::confirmed(char *msg)
{
  char str[100];
  int no, i;
  InviteData *data = (InviteData *)msg;
  BV_Member *mbr;

  if(status_ == _accepting)
    {
      no = data->numMembers;
      local_id_ = data->parts[no - 1].handle;
      BVN_DEBUG("Session confirmed, %d members, local id %d, inviting id %d\n", no, local_id_, inviting_member_->get_id());
      for(i = 0; i < no; i++)
	{
	  if(data->parts[i].host != inviting_member_->get_host())
	    {
	      mbr = new BV_Member(data->parts[i].user, data->parts[i].host);
	      mbr->set_id(data->parts[i].handle);
	      mbr->get_string(str);
	      if(data->parts[i].handle != local_id_)
		{
		  mbr->connect(reactor_);
		  BV_Msg_Header header;
		  header.set_tag(reqNewRemoteSessionConnect);
		  BV_Message msg(header);
		  mbr->send_msg(&msg);
		  
		  printf("Connect: %s, id=%d\n", str, data->parts[i].handle);
		}
	      members_[data->parts[i].handle] = *mbr;

	      mbr->get_string(str);
	      //---GUI callbacks
	      form_->addUserToList(str);
	      //---GUI callbacks END

	      delete mbr;
	    }
	  else
	    {
	      inviting_member_->set_id(data->parts[i].handle);
	      members_[inviting_member_->get_id()] = *inviting_member_;

	      inviting_member_->get_string(str);
	      //---GUI callbacks
	      form_->addUserToList(str);
	      //---GUI callbacks END

	    }

	}

      status_ = _session;
    }
  else
    return -1;

  delete inviting_member_;
  inviting_member_ = 0;
  return 0;
}

int BV_Session_Control::accepted(char *name, int host)
{
  BV_Msg_Header header;
  Participant part;

  if(new_member_ != 0)
    {
      BVN_DEBUG("BV_Session_Control::accepted\n");
      last_id_++;
      new_member_->set_id(last_id_);
      new_member_->set_name(name);

      header.set_tag(reqRemoteSessionAddMember);
      BV_Message msg0(header);
      new_member_->get_participant(&part);
      msg0.set((char *)&part, sizeof(Participant));
      send_to_all(&msg0);

      members_[last_id_] = *new_member_;

      header.set_tag(reqRemoteSessionAddRest);
      BV_Message msg(header);
      this->fill_participants(&msg);

      new_member_->send_msg(&msg);

      char str[100];
      char str1[150];
      new_member_->get_string(str);
      strcpy(str1, str);
      strcat(str1, " accepted invitation");
      //---GUI callbacks
      form_->addUserToList(str);
      form_->setStatus(str1);
      //---GUI callbacks END

      delete new_member_;
      new_member_ = 0;
    }
  else
    return -1;

  return 0;
}

int BV_Session_Control::rejected(char *name, int host)
{
  if(new_member_ != 0)
    {
      char str[100];
      char str1[150];
      new_member_->get_string(str);
      strcpy(str1, str);
      strcat(str1, " rejected invitation");
      //---GUI callbacks
      form_->setStatus(str1);
      //---GUI callbacks END

      delete new_member_;
      new_member_ = 0;
    }
  else
    return -1;

  return 0;
}

int BV_Session_Control::connect(int id, SOCK_Stream *stream)
{
  if(members_.find(id) != members_.end())
    {
      members_[id].set_stream(stream);
    }
  else
    return -1;

  return 0;
}

int BV_Session_Control::new_member(BV_Member *member)
{
  //printf("lock 1\n");
  //fflush(stdout);
  lock();

  internal_id_++;
  member->set_internal_id(internal_id_);
  members_[member->get_id()] = *member;

  if(member->get_host() == SOCK_Library::getlocalhostid())
    local_id_ = member->get_id();
    
#ifdef _DEBUG
  print_status();
#endif

  char str[100];
  char str1[150];
  member->get_string(str);
  strcpy(str1, "New user joined");
  //---GUI callbacks
  form_->getStatWnd()->getStatForm()->addToUsersList(str);
  form_->setStatus(str1);
  //---GUI callbacks END

  delete member;

  //printf("unlock 1\n");
  //fflush(stdout);
  unlock();

  return 0;
}

int BV_Session_Control::delete_member(int id)
{
  //  printf("lock 2\n");
  fflush(stdout);
  lock();

  if(members_.find(id) != members_.end())
    {
      //members_[id].close();

      char str[100];
      char str1[150];
      (*members_.find(id)).second.get_string(str);
      (*members_.find(id)).second.close_display();
      strcpy(str1, "User left");
      //---GUI callbacks
      form_->getStatWnd()->getStatForm()->delFromUsersList(str);
      //form_->delUserFromList(str);
      form_->setStatus(str1);
      //---GUI callbacks END

      members_.erase(id);

    }
  else
    {
      //  printf("unlock 2.1\n");
  fflush(stdout);
      unlock();
      return -1;
    }

  //  printf("unlock 2\n");
  fflush(stdout);
  unlock();

  return 0;
}

int BV_Session_Control::connection_closed(SOCK_Stream *stream)
{
  map <int, BV_Member>::iterator i = members_.begin();
  while(i != members_.end())
    {
      if((*i).second.get_stream() == stream)
	{

	  char str[100];
	  char str1[150];
	  (*i).second.get_string(str);
	  strcpy(str1, "User left");
	  //---GUI callbacks
	  form_->delUserFromList(str);
	  form_->setStatus(str1);
	  //---GUI callbacks END

	  members_.erase(i);
	  return 0;
	}
      i++;
    }  
  if((inviting_member_ != 0) && (inviting_member_->get_stream() == stream))
    {
      char str1[150];
      strcpy(str1, "Inviting user left");
      //---GUI callbacks
      form_->setStatus(str1);
      form_->activateButtons(False);
      //---GUI callbacks END

      delete inviting_member_;
      inviting_member_ = 0;
      status_ = _init;
      return 0;
    }

  if((new_member_ != 0) && (new_member_->get_stream() == stream))
    {
      char str1[150];
      strcpy(str1, "Invited user left");
      //---GUI callbacks
      form_->setStatus(str1);
      //---GUI callbacks END

      delete new_member_;
      new_member_ = 0;
      return 0;
    }

  return -1;
}

void BV_Session_Control::app_message(int user_id, char *msg, int len)
{
  AppData *data = (AppData *)msg;
  int size = len - sizeof(AppData) + 1;
  
  switch(data->port)
    {
    case 6666:
      if(members_.find(user_id) != members_.end())
	{
	  form_->chatMessage(&members_[user_id], data->tag, data->data, size);
	}
      break;

    default:
      break;
    }
}

////////////////////////////////////////////////////////////////////
//
// User requests
//
int BV_Session_Control::invite(char *name, char *host)
{
  int host_id;
  char lname[50];

  if((status_ != _invited) && (new_member_ == 0))
    {
      if(status_ == _init)
	{
	  BV_Member *tmp = new BV_Member(options_.local_name, SOCK_Library::getlocalhostid());
	  tmp->set_id(0);
	  members_[0] = *tmp;
	  
	  char str[100];
	  tmp->get_string(str);
	  //---GUI callbacks
	  form_->addUserToList(str);
	  //---GUI callbacks END

	  delete tmp;
	  local_id_ = 0;
	  last_id_ = 0;
	  status_ = _session;
	}

      if((host_id = SOCK_Library::gethostid(host)) == 0)
	{
	  return -1;
	}

      new_member_ = new BV_Member(name, host_id);
      if(new_member_->connect(reactor_) == -1)
	{
	  BVN_DEBUG("BV_Session_Control::invite user not present\n");
	  return -1;
	}

      BV_Msg_Header header;
      header.set_tag(reqNewRemoteSessionInvite);

      BV_Message msg(header);
      this->fill_participants(&msg);

      new_member_->send_msg(&msg);
    }

  return 0;
}

int BV_Session_Control::accept_invitation()
{
  if(status_ == _invited)
    {
      BV_Msg_Header header;
      header.set_tag(reqRemoteSessionAccept);
      BV_Message msg(header);
      inviting_member_->send_msg(&msg);
      status_ = _accepting;
      
      return 0;
    }

  return -1;
}

int BV_Session_Control::reject_invitation()
{
  if(status_ == _invited)
    {
      BV_Msg_Header header;
      header.set_tag(reqRemoteSessionReject);
      BV_Message msg(header);
      inviting_member_->send_msg(&msg);

      delete inviting_member_;
      inviting_member_ = 0;
      status_ = _init;
   
      return 0;
    }

  return -1;
}

int BV_Session_Control::quit()
{
  map <int, BV_Member>::iterator i = members_.begin();
  while(i != members_.end())
    {
      (*i).second.close();
      members_.erase(i);
      i++;
    }
  status_ = _init;
  return 0;
}

int BV_Session_Control::reset()
{
  map <int, BV_Member>::iterator i = members_.begin();
  while(i != members_.end())
    {
      //(*i).second.close();
      members_.erase(i);
      i++;
    }
  status_ = _init;
  return 0;
}


int BV_Session_Control::send_to_app(int no, int tag, char *msg, int len)
{
  AppData *data;
  int size;
  
  size = len + sizeof(AppData) - 1;
  data = (AppData *)new char[size];
  if(len > 0)
    memcpy(data->data, msg, len);
  data->tag = htons(tag);
  data->port = htonl(no);
  
  BV_Msg_Header header;
  header.set_tag(reqRemoteSessionSendToApp);
  BV_Message mesg(header);
  mesg.set((char *)data, size);
  send_to_all(&mesg);
  
  delete data;

  return 0;
}

//---GUI callbacks
void BV_Session_Control::setGUI(MainWindowMainWindow *wnd, MainForm *form)
{
  main_wnd_ = wnd;
  form_ = form;
}

MainForm *BV_Session_Control::getMainForm()
{
  return form_;
}
//---GUI callbacks END


Options *BV_Session_Control::get_options()
{
  return &options_;
}

void BV_Session_Control::update_options()
{
  main_wnd_->updateOptions(&options_);
}

int BV_Session_Control::get_local_id()
{
  return local_id_;
}

map <int, BV_Member> *BV_Session_Control::get_member_list()
{
  return &members_;
}

BV_Member *BV_Session_Control::get_member(int id)
{
  map <int, BV_Member>::iterator i = members_.find(id);
  if(i != members_.end())
    {
      return &(*i).second;
    }

  return NULL;
}

void BV_Session_Control::update_audio_stat(BV_Member *mbr)
{
  
  char str[100];
  char *str1;
  static int i = 0;

  mbr->get_string(str);
  str1 = form_->getStatWnd()->getStatForm()->getSelectedPart();

  if(strcmp(str, str1) == 0)
    {
      if(i == 0)
	form_->getStatWnd()->getStatForm()->
	  setAudioData(mbr->get_audio_data());  
      i = (i + 1) % 5;
    }
}

void BV_Session_Control::update_video_stat(BV_Member *mbr)
{
   
  char str[100];
  char *str1;
  static int i = 0;

  mbr->get_string(str);
  str1 = form_->getStatWnd()->getStatForm()->getSelectedPart();

  if(strcmp(str, str1) == 0)
    {
      if(i == 0)
	form_->getStatWnd()->getStatForm()->
	  setVideoData(mbr->get_video_data());
      i = (i + 1) % 3;
    }
}

void BV_Session_Control::handle_member_events()
{
  //printf("lock 2\n");
  //fflush(stdout);
  lock();

  map <int, BV_Member>::iterator i = members_.begin();
  while(i != members_.end())
    {
      (*i).second.handle_events();
      //update_audio_stat(&(*i).second);
      //update_video_stat(&(*i).second);
      i++;
    }
  //form_->updateLevels();

  //printf("unlock 2\n");
  //fflush(stdout);
  unlock();
}

void BV_Session_Control::update_stat(char *user)
{
  char str[100];

  map <int, BV_Member>::iterator i = members_.begin();
  while(i != members_.end())
    {
      (*i).second.get_string(str);
      if(strcmp(str, user) == 0)
	{
	  form_->getStatWnd()->getStatForm()->
	    setAudioData((*i).second.get_audio_data());  
	  form_->getStatWnd()->getStatForm()->
	    setVideoData((*i).second.get_video_data());
	}
      i++;
    }
}

void BV_Session_Control::lock()
{
  sem_wait(&sem_);  
}

void BV_Session_Control::unlock()
{
  sem_post(&sem_);  
}


//////////////////////////////////////////////////////////////////////
//
// Protected methods
//
void BV_Session_Control::fill_participants(BV_Message *msg)
{
  InviteData *data;
  int size;
  int no = 0;

  size = sizeof(InviteData) + sizeof(Participant)*members_.size() - 1;
  data = (InviteData *)new char[size];
  data->numMembers = members_.size();
  data->handle = local_id_;

  map <int, BV_Member>::iterator i = members_.begin();
  while(i != members_.end())
    {
      (*i).second.get_participant(&data->parts[no]);
      i++;
      no++;
    }

  msg->set((char *)data, size);
}

void BV_Session_Control::send_to_all(BV_Message *msg)
{
  map <int, BV_Member>::iterator i = members_.begin();
  while(i != members_.end())
    {
      char str[100];
      (*i).second.get_string(str);
      printf("%s, id=%d\n", str, (*i).second.get_id());
      if((*i).second.get_id() != local_id_)
	 (*i).second.send_msg(msg);
      i++;
    }
  
}

int BV_Session_Control::restore_options()
{
  FILE *ofile;
  char path[200];
  char *home = getenv("HOME");
  strcpy(path, home);
  strcat(path, "/.bv.cfg");
  
  ofile = fopen(path, "rb");
  if(ofile != NULL)
    {
      int no = fread(&options_, 1, sizeof(Options), ofile);
      if(no < sizeof(Options))
	return -1;
    }
  else
    return -1;

  return 0;  
}

int BV_Session_Control::save_options()
{
  FILE *ofile;
  char path[200];
  char *home = getenv("HOME");
  strcpy(path, home);
  strcat(path, "/.bv.cfg");
  
  ofile = fopen(path, "wb");
  if(ofile != NULL)
    {
      int no = fwrite(&options_, 1, sizeof(Options), ofile);
    }
  else
    return -1;

  return 0;
}

void BV_Session_Control::print_status()
{
  char str[100];
  map <int, BV_Member>::iterator i = members_.begin();

  printf("BV_Session Status\n---------------------\n");
  printf("\tLocal id: %d\n", local_id_);

  while(i != members_.end())
    {
      (*i).second.get_string(str);
      printf("Member: id %d, %s, internal id %d\n", 
	     (*i).second.get_id(), str, (*i).second.get_internal_id());
      i++;
    }

  fflush(stdout);
  
}
